#ifndef _JS_IMAGE_INFO_H
#define _JS_IMAGE_INFO_H

#include <vector>

#include "js_typedefs.h"

namespace js
{

class image_component;
class tile_component;

class sizClass
{
	public:
		sizClass();

		uint32 position;

		uint16 profile; // (Rsiz) 0, 1, or 2. Indicates what capabilities needs to be decoded.
		uint16 nComponents; // Number of components in the image

		// Sizes
		uint32 x; // Width of reference grid
		uint32 y; // Height of reference grid
		uint32 xo; // Horizontal offset (to the left side of image area)
		uint32 yo; // Vertical offset (to the top side of image area)
		uint32 xt; // Width of one reference til width respect to reference grid
		uint32 yt; // Height of one reference til width respect to reference grid
		uint32 xto; // Horizontal offset from the origin of reference grid to left side of first tile
		uint32 yto; // Vertical offset from the origin of reference grid to left side of first tile
};
class sotClass
{
	public:
		sotClass();

		uint32 position;

		uint16 tileIndex; // Index of the mother tile
		uint8 tilePartIndex; // If only one tile-part, this is 0 (appears in order, but may be interleaved with other tiles)
		uint32 tilePartLength; // If 0, the bitstream for this tile ends in EOC
		uint8 nTileParts; // Number of tile-parts of a tile in the codestream. 0 means not specified in this tile-part
};
class codClass
{
	public:
		codClass();

		bool useMe;
		uint32 position;

		/* Parameters gotten from COD marker */
		uint8 codeStyle; // SOP and EPH use? Precinct sizes predefined or custom? The precinct sizes may be overridden for each component.
		
		uint8 progressionOrder; // What kind of progression is used
		uint16 nLayers; // Number of layers
		uint8 multiCompTransform; // 1 if using multiple component transform, 0 if not specified
		
		uint8 nLevels; // Number of decomposition levels, 0 if no transform
		uint8 codeBlockWidth; // Exponent offset value, xcb
		uint8 codeBlockHeight; // Exponent offset value, ycb
		uint8 codeBlockStyle; // Style of code-block coding passes
		uint8 transform; // Wavelet transform used
		
		std::vector<uint8> precinctSizes; // Each value corresponds to a subband resolution level. Vector only in use if defined in codeStyle variable.
};
class cocClass
{
	public:
		cocClass();

		bool useMe;
		uint32 position;

		uint8 codeStyle; // 1 if custom precinctsizes

		uint8 nLevels; // Number of decomposition levels, 0 if no transform
		uint8 codeBlockWidth; // Exponent offset value, xcb
		uint8 codeBlockHeight; // Exponent offset value, ycb
		uint8 codeBlockStyle; // Style of code-block coding passes
		uint8 transform; // Wavelet transform used
		
		std::vector<uint8> precinctSizes; // Each value corresponds to a subband resolution level. Vector only in use if defined in codeStyle variable.
};
class qcdClass
{
	public:
		qcdClass();

		bool useMe;
		uint32 position;

		uint8 quantStyle; // Quantization style

		std::vector<uint16> quantStepSizes; // Actual value may be 8 bit as well, in the case of no quantization. One for each sub-band. The size should equal the greatest number of decomp. levels found in the tiles-components.
};
class qccClass
{
	public:
		qccClass();

		bool useMe;
		uint32 position;

		uint8 quantStyle; // Quantization style

		std::vector<uint16> quantStepSizes; // Actual value may be 8 bit as well, in the case of no quantization. One for each sub-band. The size should equal the greatest number of decomp. levels found in the tiles-components.
};
class rgnClass
{
	public:
		rgnClass();

		bool useMe;
		uint32 position;

		uint8 roiStyle;
		uint8 roiValue;
};
class pocClass
{
	public:
		pocClass();

		bool useMe;
		uint32 position;

		std::vector<uint8> resLevelIndexStart;
		std::vector<uint16> compIndexStart; // Value may only be 8 bit if number of comps is < 257
		std::vector<uint16> layerIndexEnd; // (starts at 0)
		std::vector<uint8> resLevelIndexEnd;
		std::vector<uint16> compIndexEnd; // Value may be 8 bit depending on number of comps
		std::vector<uint8> progr;
};
class ppmClass
{
	public:
		ppmClass();

		bool useMe;
		uint32 position;

		uint8 index;

		std::vector<uint32> infoSize;
		std::vector<uint8> packetHeader; // Should be a separate class?
};
class tlmClass
{
	public:
		tlmClass();

		bool useMe;
		uint32 position;

		uint8 index;
		uint8 size;

		std::vector<uint16> tilepartIndex; // may be 8 or 0 bits, depending on ST
		std::vector<uint32> tilePartLength; // May be 16 bits, depending on SP
};
class plmClass
{
	public:
		plmClass();

		bool useMe;
		uint32 position;

		uint8 index;

		std::vector<uint8> infoSize;
		std::vector<uint8> packetLength; // variable size!
};
class crgClass
{	
	public:
		crgClass();

		bool useMe;
		uint32 position;

		std::vector<uint16> xo;
		std::vector<uint16> yo;
};
class comClass
{
	public:
		comClass();

		bool useMe;
		uint32 position;

		uint16 registration;
		std::vector<uint8> data;
};
class pptClass
{
	public:
		pptClass();

		bool useMe;
		uint32 position;

		uint8 index;

		std::vector<uint8> packetHeader; // Should be a separate class?
};
class pltClass
{
	public:
		pltClass();

		bool useMe;
		uint32 position;

		uint8 index;

		std::vector<uint8> packetLength; // Variable length;
};

class image
{
	public:
		image();

		/* Length of this header in bytes */
		uint32 headerLength;

		/* Component specific information goes here */
		std::vector<image_component> comps; 

		sizClass siz; // Sizes and number of comps
		codClass cod; // Coding style
		qcdClass qcd; // Quantization style
		pocClass poc; // Progr change
		ppmClass ppm; // Packed packet header
		crgClass crg; // Comp registration

		/* Markers there may be several of */
		std::vector<tlmClass> tlms;
		std::vector<plmClass> plms;
		std::vector<comClass> coms; // Comments
};
class image_component
{
	public:
		image_component();

		cocClass coc; // Coding style component
		qccClass qcc; // Quantization style component
		rgnClass rgn; // Region of interest

		uint8 precision;
		uint8 horizontalSeparation; // wrt. reference grid
		uint8 verticalSeparation;		
};

class tile
{
	public:
		tile();

		/* Length of this header in bytes */
		uint32 headerLength; 

		/* Length of the bitstream after the header. Includes SOD marker! */
		uint32 bitstreamLength;

		/* Component specific information */ 
		std::vector<tile_component> comps;

		/* Positions of the different layers for this tile part */
		std::vector<std::pair<uint32,uint32> > layers;

		sotClass sot; // start of tile part header, size, lengths and indices
		codClass cod; // Coding style
		qcdClass qcd; // Quantization style
		pocClass poc; // Progr change
		pptClass ppt; // Packed packet header, tile

		/* Markers there may be several of */
		std::vector<pltClass> plts;
		std::vector<comClass> coms;		
};
class tile_component
{
	public:
		tile_component();

		cocClass coc; // Coding style component
		qccClass qcc; // Quantization style component
		rgnClass rgn; // Region of interest
};

} // End namespace js

#endif