#include <iostream>

#include "js_typedefs.h"
#include "js_parse_funcs.h"
#include "js_useful_funcs.h"
#include "js_constants.h"
#include "js_image_info.h"

namespace js
{

/* Generic parsing functions to use with the parse_state class */
exit_res scale(char* image_buf, uint32 nBuffer, uint32 &nBufferNew, uint32 target, uint16 targetLayers, bool human)
{

	if(target >= nBuffer) // If you want more than what is in the picture...
	{
		nBufferNew = nBuffer;
		return exit_success; 
	}

	 parse_state* p = new parse_state(image_buf, nBuffer, human);

	 if(p->parseMainHeader() != exit_success){
		 if(p->human) std::cout << "Parse of main header did not finish as expected." << std::endl;
		 return exit_fail;
	 }

	 /* Start reading tile-headers and bitstreams until the end of the codestream */
	 while(p->lastMarker != marker::EOC)
	 {
		 // Current marker should be SOT
		if(p->parseTilePartHeader() != exit_success){
			if(p->human) std::cout << "Parse of tile-part header did not finish as expected." << std::endl;
			return exit_fail;
		}
		// Current marker should be SOD
		if(p->parseBitstream() != exit_success){
			if(p->human) std::cout << "Parse of bitstream did not finish as expected." << std::endl;
			return exit_fail;
		}
		// Current marker should be either EOC or SOT
	 }

	 p->scaleImage(target, targetLayers, nBufferNew);
	 
	 delete p;
	 return exit_success;
}

/* Parse_state functions */
parse_state::parse_state(char* image_buf, uint32 sizeOfBuffer, bool human_in)
{
	nBuffer = sizeOfBuffer;
	buffer = image_buf;
	bufferPosition = 0;
	lastMarker = 0;
	human = human_in;
	in_main = true;
}
exit_res parse_state::parseMainHeader()
{
	in_main = true;
	uint16 currMarker;

	// Read the buffer until we find SOC (start of codestream) marker
	do
	{
		if(readNextMarker(currMarker) != exit_success) return exit_fail;
	}while(currMarker != marker::SOC);
	lastMarker = marker::SOC;

	uint32 startPos = bufferPosition;

	// Scan for markers until we find the SOT (start of tile-part header) marker
	// Process the info contained in each marker segment along the way
	while(true)
	{
		if(readNextMarker(currMarker) != exit_success) return exit_fail;
		exit_res result;

		switch(currMarker)
		{
			case marker::SIZ:
				result = processSIZ();
				break;
			case marker::COD:
				result = processCOD();
				break;
			case marker::COC:
				result = processCOC();
				break;
			case marker::QCD:
				result = processQCD();
				break;
			case marker::QCC:
				result = processQCC();
				break;
			case marker::RGN:
				result = processRGN();
				break;
			case marker::POC:
				result = processPOC();
				break;
			case marker::PPM:
				result = processPPM();
				break;
			case marker::TLM:
				result = processTLM();
				break;
			case marker::PLM:
				result = processPLM();
				break;
			case marker::CRG:
				result = processCRG();
				break;
			case marker::COM:
				result = processCOM();
				break;
			case marker::SOT: // Start of tile part header means the main header is finished
				header_main.headerLength = bufferPosition - startPos; // We count the SOT marker, but at the same time we didn't count the SOC marker in the beginning, so it's correct.
				lastMarker = marker::SOT;
				return exit_success;
			case marker::EOC:
				lastMarker = marker::EOC;
				if(human) std::cout << "End of codestream reached unexpectedly." << std::endl;
				return exit_fail;
			default:
				result = processUNKNOWN(currMarker);
				break;
		}
		if(result != exit_success) return exit_fail;
	}
	return exit_fail; // Should never get this far
}
exit_res parse_state::parseTilePartHeader()
{
	in_main = false;
	uint16 currMarker;

	if(lastMarker != marker::SOT)
	{
		if(human) std::cout << "SOT marker wasn't found where I expected. Scanning for it..." << std::endl;
		do
		{
			if(readNextMarker(currMarker) != exit_success) return exit_fail;
		}while(currMarker != marker::SOT);
		lastMarker = marker::SOT;
	}
	uint32 startPos = bufferPosition;

	if(processSOT() != exit_success) return exit_fail;

	// Scan for markers until we find the SOD (start of data) marker
	// Process the info contained in each marker segment along the way
	while(true)
	{
		if(readNextMarker(currMarker) != exit_success) return exit_fail;
		exit_res result;

		switch(currMarker)
		{
			if(header_tilePart.back().sot.tilePartIndex == 0) // Only the first tile-part header in a tile may contain these markers
			{
				case marker::COD:
					result = processCOD();
					break;
				case marker::COC:
					result = processCOC();
					break;
				case marker::QCD:
					result = processQCD();
					break;
				case marker::QCC:
					result = processQCC();
					break;
				case marker::RGN:
					result = processRGN();
					break;
			}
			case marker::POC:
				result = processPOC();
				break;
			case marker::PPT:
				result = processPPT();
				break;
			case marker::PLT:
				result = processPLT();
				break;
			case marker::COM:
				result = processCOM();
				break;
			case marker::SOD: // Start of data header means the tile-part header is finished
				header_tilePart.back().headerLength = bufferPosition - startPos;  // We count the SOD marker, but at the same time we didn't count the SOT marker in the beginning, so it's correct.
				if(header_tilePart.back().sot.tilePartLength != 0)
				{
					header_tilePart.back().bitstreamLength = header_tilePart.back().sot.tilePartLength - header_tilePart.back().headerLength; 
				}
				lastMarker = marker::SOD;
				return exit_success;
			case marker::EOC:
				lastMarker = marker::EOC;
				if(human) std::cout << "End of codestream reached unexpectedly." << std::endl;
				return exit_fail;
			default:
				result = processUNKNOWN(currMarker);
				break;
		}
		if(result != exit_success) return exit_fail;
	}
	return exit_fail; // Should never get this far
}
exit_res parse_state::parseBitstream()
{
	// If length=0, then this is the last tile, and it ends with the EOC marker
	// If length!=0, then it specifies the length of the bitstream.
	// If length!=0 AND it is the last tile, then two bytes will be left in the end, which is the EOC marker. This is handled in the tile parser and parser functions.

	/* Length of the bitstream (including SOD). Subtract 2 since the SOD is already read. */
	uint32 length = header_tilePart.back().bitstreamLength - 2; 

	/* The first position of the bitstream (after SOD, that is) */
	uint32 startPos = bufferPosition; 

	/* Position of the next byte after the bitstream */
	uint32 endPos = length+startPos; 
	
	uint8 progr = 0;
	bool sop = false;

	if(header_tilePart.back().cod.useMe){
		progr = header_tilePart.back().cod.progressionOrder;
		if((header_tilePart.back().cod.codeStyle & 2) == 2) sop = true;
	}else{ 
		progr = header_main.cod.progressionOrder;
		if((header_main.cod.codeStyle & 2) == 2) sop = true;
	}

	if(progr == progression::LRCP && sop){
		return processLRCPsop();
	}else if(progr == progression::LRCP && !sop){
		return processLRCP();
	}else if(progr == progression::CPRL && sop){
		return processCPRLsop();
	}else if(progr == progression::CPRL && !sop){
		return processCPRL();
	}else if(progr == progression::PCRL && sop){
		return processPCRLsop();
	}else if(progr == progression::PCRL && !sop){
		return processPCRL();
	}else if(progr == progression::RLCP && sop){
		return processRLCPsop();
	}else if(progr == progression::RLCP && !sop){
		return processRLCP();
	}else if(progr == progression::RPCL && sop){
		return processRPCLsop();
	}else if(progr == progression::RPCL && !sop){
		return processRPCL();
	}else{
		std::cout << "Unknown progression." << std::endl;
		bufferPosition = endPos; // Might fail if length = 0
		return exit_fail;
	}
}
exit_res parse_state::scaleImage(uint32 target,  uint16 targetLayers, uint32 &newSize)
{
	 /* TODO: Weight each tile unequally, and each tile-part within a tile equally.
			  This might require writing a new COD to the tile-part header, which will be a pain.
	 */
	 
	 /* Finding the correct number of layers from the target number of bytes */
	 /* Currently we assume that the layer info is not updated in the tile headers */

	 uint32 nTilePts = header_tilePart.size();

	 /* Print layer sizes */
	 if(human)
	 {
		 for(uint32 j = 0; j < nTilePts; j++) 
		 {
			 std::cout << "Tile " << j << " :" << std::endl;
			 for(uint32 i = 0; i < header_tilePart[j].layers.size(); i++)
			 {
				 uint32 sz = header_tilePart[j].layers[i].second - header_tilePart[j].layers[i].first;
				 std::cout << "nLayer" << i << ": " << sz << std::endl;
			 }
			 std::cout << std::endl;
		 }
	 }

	 /* Assuming all tile parts have the same number of layers */
	 uint16 nLayers = header_main.cod.nLayers;
	 uint16 L = nLayers-1;
	 uint32 sub = 0;
	 bool ok = false;
	 uint32 tot = nBuffer;

	 if(targetLayers == 0)
	 {
		 /* Calculate the number of layers to be used */
		 while(L >= 1)
		 {
			 for(uint32 j = 0; j < nTilePts; j++) 
			 {
				 sub = header_tilePart[j].layers[L].second - header_tilePart[j].layers[L].first;
				 tot -= sub;
				 header_tilePart[j].sot.tilePartLength -= sub;
			 }

			 if(tot < target)
			 {
				 ok = true;
				 break;
			 }
			 L--;
		 }
		 if(!ok) L = 1;
	 }else{
		 L = targetLayers;
		 if(L > nLayers) L = nLayers;
		 if(L != nLayers) // Keeping all layers
		 {
			 for(uint16 i = L; i < nLayers; i++)
			 {
				 for(uint32 j = 0; j < nTilePts; j++)
				 {
					 header_tilePart[j].sot.tilePartLength -= header_tilePart[j].layers[i].second - header_tilePart[j].layers[i].first; 
				 }
			 }
		 }
	 }
	 if(human) std::cout << "Using " << L << " layers" << std::endl;
	
	 // Update main header
	 header_main.cod.nLayers = L;
	 updateCOD(header_main.cod);

	 // Update the tile-part headers as necessary
	 for(uint32 i = 0; i < nTilePts; i++)
	 {
		 header_tilePart[i].cod.nLayers = L;
		 updateCOD(header_tilePart[i].cod);
		 updateSOT(header_tilePart[i].sot);
	 }

	 uint32 w = 0;
	 bufferPosition = 0; 
	 uint32 lim;
	 uint8 temp;

	 /* Modify the image buffer so that it contains only the wanted layers.
		This size of the buffer is the same, but the output size parameter is changed */
	 /* Currently this only works for lrcp */
	 for(uint32 i = 0; i < nTilePts; i++)
	 {
		 lim = header_tilePart[i].layers[L-1].second;
		 while(bufferPosition < lim)
		 {
			 read8(temp);
			 write8(temp, w);
			 w++;
		 }
		 bufferPosition = header_tilePart[i].layers[nLayers-1].second;
	 }
	 while(bufferPosition < nBuffer)
	 {
		 read8(temp);
		 write8(temp, w);
		 w++;
	 }

	 newSize = w; // The output buffer size
	 return exit_success;
}

exit_res parse_state::readNextMarker(uint16 &marker)
{
	uint8 param8;

	do
	{
		if(read8(param8) != exit_success) return exit_fail;
	}while(param8 != marker::MSTART);

	if(read8(param8) != exit_success) return exit_fail;
	marker = bytesToWord(marker::MSTART, param8);

	return exit_success;
}

exit_res parse_state::processLRCPsop()
{
	uint16 currMarker;

	uint8 Nmax; // Max decomp levels
	uint16 Lmax; // Max number of layers
	uint16 Cmax = header_main.siz.nComponents; // Number of components
	uint32 Pmax = 1; // Number of precincts

	if(header_tilePart.back().cod.useMe){
		Nmax = header_tilePart.back().cod.nLevels;
		Lmax = header_tilePart.back().cod.nLayers;
	}else{
		Nmax = header_main.cod.nLevels;
		Lmax = header_main.cod.nLayers;
	}

	// Read until we find the first SOP marker. Indicates the start of the first packet, and first layer.
	// There should always be at least one SOP marker.
	do
	{
		if(readNextMarker(currMarker) != exit_success) return exit_fail;
		if(currMarker == marker::EOC)
		{
			if(human) std::cout << "EOC reached while searching for SOP. Are you sure you included SOP markers in the codestream?" << std::endl;
			return exit_fail;
		}
	}while(currMarker != marker::SOP);	
	lastMarker = marker::SOP;

	// Find the start position of the first layer
	std::pair<uint32, uint32> temp;
	temp.first = bufferPosition - 2; // Subtract the size of the SOP marker itself

	for(uint16 L = 0; L < Lmax; L++)
	{
		header_tilePart.back().layers.push_back(temp); // Push a new layer range onto the newest tile update object
		for(uint8 r = 0; r <= Nmax; r++)
		{
			for(uint16 i = 0; i < Cmax; i++)
			{
				Pmax = nPrecincts(r, Nmax, i, header_tilePart.size()-1); // The number of precincts

				for(uint32 k = 0; k < Pmax; k++)
				{
					do
					{
						if(readNextMarker(currMarker) != exit_success) return exit_fail;
						if(currMarker == marker::SOT || currMarker == marker::EOC)
						{
							header_tilePart.back().layers.back().second = bufferPosition - 2; // Subtract the length of the SOT/EOC marker itself
							lastMarker = currMarker;

							// If we finish prematurely, it's probably because of empty layers which doesn't have SOP markers. Can skip all of those.
							if(L != Lmax-1)
							{
								break; // Or if we read a SOT or EOC marker too early, there's probably more to read. Why does this happen?
								header_tilePart.back().cod.nLayers = L+1;
								// Update tilePart
								if(updateCOD(header_tilePart.back().cod) != exit_success) return exit_fail;
							}
							return exit_success;
						}
					}while(currMarker != marker::SOP);
				}
			}
		}
		header_tilePart.back().layers.back().second = bufferPosition - 2; // Subtract the length of the SOP marker itself
		temp.first = bufferPosition - 2;
	}

	// Whenever we get here, it seems to be because of some bug
	std::cout << "Warning: This codestream has more layers than specified!" << std::endl;

	if(header_tilePart.back().cod.useMe)
	{
		header_tilePart.back().cod.nLayers++;
		if(updateCOD(header_tilePart.back().cod) != exit_success) return exit_fail;
	}else{
		header_main.cod.nLayers++;
		if(updateCOD(header_main.cod) != exit_success) return exit_fail;
	}

	header_tilePart.back().layers.push_back(temp); 
	while(currMarker != marker::EOC && currMarker != marker::SOT)
	{
		if(readNextMarker(currMarker) != exit_success) return exit_fail;
	}
	
	header_tilePart.back().layers.back().second = bufferPosition - 2; // Subtract the length of the SOT/EOC marker itself
	lastMarker = currMarker;

	return exit_success;
}
exit_res parse_state::processLRCP()
{
	std::cout << "LRCP progression without SOP marker is not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processRLCPsop()
{
	std::cout << "RLCP progression not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processRLCP()
{
	std::cout << "RLCP progression not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processRPCLsop()
{
	std::cout << "RPCL progression not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processRPCL()
{
	std::cout << "RPCL progression not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processPCRLsop()
{
	std::cout << "PCRL progression not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processPCRL()
{
	std::cout << "PCRL progression not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processCPRLsop()
{
	std::cout << "CPRL progression not supported." << std::endl;
	return exit_fail;
}
exit_res parse_state::processCPRL()
{
	std::cout << "CPRL progression not supported." << std::endl;
	return exit_fail;
}

exit_res parse_state::updateCOD(codClass &thisCod)
{
	if(thisCod.useMe)
	{
		return write16(thisCod.nLayers, thisCod.position+2);
	}else{
		return exit_success;
	}
}
exit_res parse_state::updateSOT(sotClass &thisSot)
{
	return write32(thisSot.tilePartLength, thisSot.position+2);
}

exit_res parse_state::read8(uint8& ret)
{
	if(bufferPosition >= nBuffer)
	{
		if(human) std::cout << "Read8 failed. End of file reached." << std::endl;
		return exit_fail;
	}
	ret = buffer[bufferPosition];
	bufferPosition++;

	return exit_success;
}
exit_res parse_state::read16(uint16& ret)
{
	if(bufferPosition + 1 >= nBuffer)
	{
		if(human) std::cout << "Read16 failed. End of file reached." << std::endl;
		return exit_fail;
	}
	uint8 byte1 = buffer[bufferPosition];
	uint8 byte2 = buffer[bufferPosition+1];
	bufferPosition += 2;

	ret = uint16((byte1<<8) | byte2);
	return exit_success;
}
exit_res parse_state::read32(uint32& ret)
{
	if(bufferPosition + 3 >= nBuffer)
	{
		if(human) std::cout << "Read16 failed. End of file reached." << std::endl;
		return exit_fail;
	}
	uint8 byte1 = buffer[bufferPosition];
	uint8 byte2 = buffer[bufferPosition+1];
	uint8 byte3 = buffer[bufferPosition+2];
	uint8 byte4 = buffer[bufferPosition+3];
	bufferPosition += 4;

	ret = uint32( (byte1 << 24) | (byte2 << 16) | (byte3 << 8) | byte4);
	return exit_success;
}
exit_res parse_state::write8(uint8 data, uint32 pos)
{
	if(pos >= nBuffer) 
	{
		if(human) std::cout << "write8 fail. Out of buffer range." << std::endl;
		return exit_fail;
	}

	buffer[pos] = data;

	return exit_success;
}
exit_res parse_state::write16(uint16 data, uint32 pos)
{
	if(pos+1 >= nBuffer)
	{
		if(human) std::cout << "write16 fail. Out of buffer range." << std::endl;
		return exit_fail;
	}

	buffer[pos] = uint8((data & 0xFF00) >> 8);
	buffer[pos+1] = data & 0x00FF;

	return exit_success;
}
exit_res parse_state::write32(uint32 data, uint32 pos)
{
	if(pos+3 >= nBuffer)
	{
		if(human) std::cout << "write16 fail. Out of buffer range." << std::endl;
		return exit_fail;
	}

	buffer[pos]   = uint8((data & 0xFF000000) >> 24);
	buffer[pos+1] = uint8((data & 0x00FF0000) >> 16);
	buffer[pos+2] = uint8((data & 0x0000FF00) >> 8);
	buffer[pos+3] = uint8(data & 0x000000FF);

	return exit_success;
}

uint32 parse_state::nPrecincts(int r, int rmax, int c, int tilePartNumber)
{
	/* Something is wrong with this function when using both tiling and precinct custom value */

	// Number of precincts for resolution lvl r, and for 
	// the tile currently parsing

	uint8 ppx;
	uint8 ppy;
	uint8 pp;

	if(header_tilePart[tilePartNumber].comps[c].coc.useMe)
	{
		if((header_tilePart[tilePartNumber].comps[c].coc.codeStyle & 1) == 0) return 1;
		pp = header_tilePart[tilePartNumber].comps[c].coc.precinctSizes[r];
	}else if(header_tilePart[tilePartNumber].cod.useMe){
		if((header_tilePart[tilePartNumber].cod.codeStyle & 1) == 1) return 0;
		pp = header_tilePart[tilePartNumber].cod.precinctSizes[r];
	}else if(header_main.comps[c].coc.useMe)
	{
		if((header_main.comps[c].coc.codeStyle & 1) == 0) return 1;
		pp = header_main.comps[c].coc.precinctSizes[r];
	}else{
		if((header_main.cod.codeStyle & 1) == 0) return 1;
		pp = header_main.cod.precinctSizes[r];
	}

	ppx = pp & 0x0F;
	ppy = (pp & 0xF0) >> 4;

	uint32 numXtiles = uint32(ceil((header_main.siz.x - header_main.siz.xto) /  double(header_main.siz.xt)));
	uint32 numYtiles = uint32(ceil((header_main.siz.y - header_main.siz.yto) /  double(header_main.siz.yt)));

	uint32 pleft = header_tilePart[tilePartNumber].sot.tileIndex % numXtiles;
	uint32 qleft = uint32(ceil(header_tilePart[tilePartNumber].sot.tileIndex / double(numXtiles)));

	uint32 tx0 = std::max(header_main.siz.xto + pleft*header_main.siz.xt, header_main.siz.xo);
	uint32 ty0 = std::max(header_main.siz.yto + qleft*header_main.siz.yt, header_main.siz.yo);

	uint32 tx1 = std::max(header_main.siz.xto + (pleft+1)*header_main.siz.xt, header_main.siz.x);
	uint32 ty1 = std::max(header_main.siz.yto + (qleft+1)*header_main.siz.yt, header_main.siz.y);

	uint32 tcx0 = uint32(ceil(tx0/double(header_main.comps[c].horizontalSeparation)));
	uint32 tcx1 = uint32(ceil(tx1/double(header_main.comps[c].horizontalSeparation)));

	uint32 tcy0 = uint32(ceil(ty0/double(header_main.comps[c].verticalSeparation)));
	uint32 tcy1 = uint32(ceil(ty1/double(header_main.comps[c].verticalSeparation)));

	double div = pow(2, double(rmax - r));

	uint32 trx0 = uint32(ceil(tcx0/div));
	uint32 try0 = uint32(ceil(tcy0/div));

	uint32 trx1 = uint32(ceil(tcx1/div));
	uint32 try1 = uint32(ceil(tcy1/div));

	if(trx1 == trx0) return 0;
	if(try1 == try0) return 0;

	uint32 num2w = uint32(ceil(trx1/pow(2, double(ppx)))) - uint32(ceil(trx0/pow(2, double(ppx))));
	uint32 num2h = uint32(ceil(try1/pow(2, double(ppy)))) - uint32(ceil(try0/pow(2, double(ppy))));

	return num2w*num2h;
}

/* Marker segment processing functions */
exit_res parse_state::processSOT()
{
	lastMarker = marker::SOT;
	
	tile temp;

	uint16 length;
	uint8 param8;
	uint16 param16;
	uint32 param32;

	if(read16(length) != exit_success) return exit_fail;
	temp.sot.position = bufferPosition;	

	if(read16(param16) != exit_success) return exit_fail;
	temp.sot.tileIndex = param16;
	
	if(read32(param32) != exit_success) return exit_fail;
	temp.sot.tilePartLength = param32;  // See figure A.16 for this. The length includes the entire SOP marker and segment (and the entire SOD marker, and the entire tile-part marker segment length, but this we remove at the end of the tile-part parser)

	if(read8(param8) != exit_success) return exit_fail;
	temp.sot.tilePartIndex = param8;

	if(read8(param8) != exit_success) return exit_fail;
	temp.sot.nTileParts = param8;

	/* Add tile components */
	temp.comps.reserve(header_main.siz.nComponents);
	for(uint32 i = 0; i < header_main.siz.nComponents; i++)
	{
		tile_component temp2;		
		temp.comps.push_back(temp2);
	}

	header_tilePart.push_back(temp);

	return exit_success;
}
exit_res parse_state::processSIZ()
{
	lastMarker = marker::SIZ;

	uint16 length;
	uint8 param8;
	uint16 param16;
	uint32 param32;

	if(read16(length) != exit_success) return exit_fail;
	header_main.siz.position = bufferPosition;
	
	if(read16(param16) != exit_success) return exit_fail;
	header_main.siz.profile = param16;

	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.x = param32;

	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.y = param32;
	
	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.xo = param32;

	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.yo = param32;

	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.xt = param32;

	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.yt = param32;

	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.xto = param32;

	if(read32(param32) != exit_success) return exit_fail;
	header_main.siz.yto = param32;

	if(read16(param16) != exit_success) return exit_fail;
	header_main.siz.nComponents = param16; 

	header_main.comps.reserve(header_main.siz.nComponents);

	for(uint16 i = 0; i < header_main.siz.nComponents; i++)
	{
		/* Add image component */

		image_component temp;
		
		if(read8(param8) != exit_success) return exit_fail;
		temp.precision = param8;

		if(read8(param8) != exit_success) return exit_fail;
		temp.horizontalSeparation = param8;

		if(read8(param8) != exit_success) return exit_fail;
		temp.verticalSeparation = param8;

		header_main.comps.push_back(temp);
	}

	return exit_success;
}

exit_res parse_state::processCOD()
{
	lastMarker = marker::COD;

	codClass *cod = &(header_main.cod);
	if(!in_main) cod = &(header_tilePart.back().cod);

	cod->useMe = true;

	// The parameters set in this marker segment can be overridden by COC for an individual comp either in main or tile-part header
	uint16 length;
	uint8 param8;
	uint16 param16;

	if(read16(length) != exit_success) return exit_fail;
	cod->position = bufferPosition;

	if(read8(param8) != exit_success) return exit_fail;
	cod->codeStyle = param8;

	if(read8(param8) != exit_success) return exit_fail;
	cod->progressionOrder = param8;

	if(read16(param16) != exit_success) return exit_fail;
	cod->nLayers = param16;

	if(read8(param8) != exit_success) return exit_fail;
	cod->multiCompTransform = param8;

	if(read8(param8) != exit_success) return exit_fail;
	cod->nLevels = param8;

	if(read8(param8) != exit_success) return exit_fail;
	cod->codeBlockWidth = param8;

	if(read8(param8) != exit_success) return exit_fail;
	cod->codeBlockHeight = param8;

	if(read8(param8) != exit_success) return exit_fail;
	cod->codeBlockStyle = param8;

	if(read8(param8) != exit_success) return exit_fail;
	cod->transform = param8;

	if(!(cod->codeStyle & 1)) return exit_success; // If default precinct sizes
	if(cod->precinctSizes.size() != 0) cod->precinctSizes.resize(0);

	cod->precinctSizes.reserve(1 + cod->nLevels); // The number of resolution levels equals the number of decomposition levels + 1

	for(int i = 0; i <= cod->nLevels; i++)
	{
		if(read8(param8) != exit_success) return exit_fail;
		cod->precinctSizes.push_back(param8);
	}

	return exit_success;
}
exit_res parse_state::processCOC()
{	
	lastMarker = marker::COC;

	uint16 length;
	uint16 index;
	uint8 param8;

	read16(length);
	
	if(header_main.siz.nComponents >= 257)
	{
		read16(index);
	}else{
		read8(param8);
		index = uint16(param8);
	}

	cocClass *coc = &(header_main.comps[index].coc);
	if(!in_main) coc = &(header_tilePart.back().comps[index].coc);

	coc->useMe = true;
	coc->position = bufferPosition;

	read8(param8);
	coc->codeStyle = param8; // Only contains precinctinfo

	read8(param8);
	coc->nLevels = param8;

	read8(param8);
	coc->codeBlockWidth = param8;

	read8(param8);
	coc->codeBlockHeight = param8;

	read8(param8);
	coc->codeBlockStyle = param8;

	read8(param8);
	coc->transform = param8;

	if(coc->codeStyle == 0) return exit_success;
	if(coc->precinctSizes.size() != 0) coc->precinctSizes.resize(0);

	coc->precinctSizes.reserve(1 + coc->nLevels); 

	for(int i = 0; i <= coc->nLevels; i++)
	{
		read8(param8);
		coc->precinctSizes.push_back(param8);
	}

	return exit_success;
}
exit_res parse_state::processQCD()
{
	lastMarker = marker::QCD;

	qcdClass *qcd = &(header_main.qcd);
	if(!in_main) qcd = &(header_tilePart.back().qcd);

	qcd->useMe = true;

	uint16 length;
	uint8 param8;
	uint16 param16;
	
	read16(length);
	length -= 2;

	qcd->position = bufferPosition;

	read8(param8);
	qcd->quantStyle = param8;
	length--;

	// Length is now the remaining number of bytes

	param8 <<= 3; // Shifting 3 bits to the right
	param8 >>= 3; // and back again, to get rid of the 3 msb's

	if(qcd->quantStepSizes.size() != 0) qcd->quantStepSizes.resize(0);
	
	if(param8 == 1 || param8 == 2) // The quantization step sizes are 16 bit values
	{
		qcd->quantStepSizes.reserve(length/2);
		for(uint16 i = 0; i < length/2; i++)
		{
			read16(param16);
			qcd->quantStepSizes.push_back(param16);
		}
	}else if(param8 == 0){ // The quantization exponent is signaled (which is 8 bit)
		qcd->quantStepSizes.reserve(length);

		for(uint16 i = 0; i < length; i++)
		{
			read8(param8);
			qcd->quantStepSizes.push_back(param8);
		}
	}

	return exit_success;
}
exit_res parse_state::processQCC()
{
	lastMarker = marker::QCC;

	uint16 length;
	uint16 index;
	uint8 param8;
	uint16 param16;

	read16(length);
	length -= 2;
	
	if(header_main.siz.nComponents >= 257)
	{
		read16(index);
		length -= 2;
	}else{
		read8(param8);
		index = uint16(param8);
		length--;
	}

	qccClass *qcc = &(header_main.comps[index].qcc);
	if(!in_main) qcc = &(header_tilePart.back().comps[index].qcc);

	qcc->useMe = true;
	qcc->position = bufferPosition;

	read8(param8);
	qcc->quantStyle = param8;
	length--;

	// Length is now the number of bytes that is left in this marker segment

	param8 <<= 3; // Shifting 3 bits to the right
	param8 >>= 3; // and back again, to get rid of the 3 msb's

	if(qcc->quantStepSizes.size() != 0) qcc->quantStepSizes.resize(0);
	
	if(param8 == 1 || param8 == 2) // The quantization step sizes are 16 bit values
	{
		qcc->quantStepSizes.reserve(length/2);
		for(uint16 i = 0; i < length/2; i++)
		{
			read16(param16);
			qcc->quantStepSizes.push_back(param16);
		}
	}else if(param8 == 0){ // The quantization exponent is signaled (which is 8 bit)
		qcc->quantStepSizes.reserve(length);

		for(uint16 i = 0; i < length; i++)
		{
			read8(param8);
			qcc->quantStepSizes.push_back(param8);
		}
	}

	return exit_success;
}
exit_res parse_state::processRGN()
{
	lastMarker = marker::RGN;

	uint16 length;
	uint16 index;
	uint8 param8;

	read16(length);
	length -= 2;
	
	if(header_main.siz.nComponents >= 257)
	{
		read16(index);
		length -= 2;
	}else{
		read8(param8);
		index = uint16(param8);
		length--;
	}

	rgnClass *rgn = &(header_main.comps[index].rgn);
	if(!in_main) rgn = &(header_tilePart.back().comps[index].rgn);

	rgn->useMe = true;
	rgn->position = bufferPosition;

	read8(param8);
	rgn->roiStyle = param8;

	read8(param8);
	rgn->roiValue = param8;

	return exit_success;
}
exit_res parse_state::processPOC()
{
	if(human) std::cout << "Progresssion order change detected. This feature is not supported yet." << std::endl;
	return exit_fail;

	lastMarker = marker::POC;
	return(processUNKNOWN(marker::POC));
}

exit_res parse_state::processPLM()
{
	if(human) std::cout << "Pointer marker segments are not supported." << std::endl;
	return exit_fail;

	lastMarker = marker::PLM;
	return(processUNKNOWN(marker::PLM));
}
exit_res parse_state::processPLT()
{
	if(human) std::cout << "Pointer marker segments are not supported." << std::endl;
	return exit_fail;

	lastMarker = marker::PLT;
	return(processUNKNOWN(marker::PLT));
}
exit_res parse_state::processPPM()
{
	if(human) std::cout << "Pointer marker segments are not supported." << std::endl;
	return exit_fail;

	lastMarker = marker::PPM;
	return(processUNKNOWN(marker::PPM));
}
exit_res parse_state::processPPT()
{
	if(human) std::cout << "Pointer marker segments are not supported." << std::endl;
	return exit_fail;

	lastMarker = marker::PPT;
	return(processUNKNOWN(marker::PPT));
}
exit_res parse_state::processTLM()
{
	if(human) std::cout << "Pointer marker segments are not supported." << std::endl;
	return exit_fail;

	lastMarker = marker::TLM;
	return(processUNKNOWN(marker::TLM));
}

exit_res parse_state::processCRG()
{
	lastMarker = marker::CRG;
	return(processUNKNOWN(marker::CRG));
}
exit_res parse_state::processCOM()
{
	lastMarker = marker::COM;
	return(processUNKNOWN(marker::COM));
}

exit_res parse_state::processUNKNOWN(uint16 marker)
{
	if(human) std::cout << "Marker segment of type (" << markerToString(marker) << ") discarded." << std::endl;
	lastMarker = marker;

	// Use length to discard the rest of the marker (actually move the position, not discard).
	uint16 length;
	read16(length);
	
	// lsiz is the size of the marker segment including lsiz itself
	bufferPosition += length-2; // lsiz is 2 bytes itself
	
	return exit_success;
}

} // End namespace js