#ifndef _JS_PARSE_FUNCS_H
#define _JS_PARSE_FUNCS_H

#include "js_image_info.h"

namespace js
{

exit_res scale(char* buffer, uint32 nBuffer, uint32 &nBufferNew, uint32 target, uint16 targetLayers, bool human);

/* Class for keeping track of headers and image buffer */
class parse_state
{
	private:
		/* Image buffer */
		char* buffer;

		/* Buffer state */
		uint32 nBuffer; // Size of image buffer
		uint32 bufferPosition; // Between 0 and nBuffer-1

	public:
		/* Constructor */
		parse_state(char* image_buf, uint32 sizeOfBuffer, bool human_in);
		
		/* The most recently encountered marker */
		uint16 lastMarker;
	
		/* Is a human using the program? If so, we print out debug comments. */
		bool human;

		/* Parse headers and bitstreams */
		exit_res parseMainHeader();
		exit_res parseTilePartHeader();
		exit_res parseBitstream();

		 /* Used the information gathered by the image to scale the image, aiming for the target size */
		exit_res scaleImage(uint32 target,  uint16 targetLayers, uint32 &newSize);

	private:
		/* Are we currently parsing the main header? */
		bool in_main;
		
		/* Image information gotten from main header */
		image header_main;

		/* Information about the tile parts (and bitstreams inside those tile parts) */
		std::vector<tile> header_tilePart;

		/* Read and write to buffer, big endian style */
		exit_res read8(uint8& ret);
		exit_res read16(uint16& ret);
		exit_res read32(uint32& ret);
		exit_res write8(uint8 data, uint32 pos);
		exit_res write16(uint16 data, uint32 pos);
		exit_res write32(uint32 data, uint32 pos);

		/* Returns number of precincts in resolution lvl r, component c, and tile part index */
		uint32 nPrecincts(int r, int rmax, int c, int tilePartNumber);

		/* Read the next marker in the codestream */
		exit_res readNextMarker(uint16 &marker);

		/* Layer - resolution - component - position */
		exit_res processLRCPsop();
		exit_res processLRCP();

		/* Resolution - layer - component - position */
		exit_res processRLCPsop();
		exit_res processRLCP();

		/* Resolution - position - component - layer */
		exit_res processRPCLsop();
		exit_res processRPCL();

		/* Position - component - resolution - layer */
		exit_res processPCRLsop();
		exit_res processPCRL();

		/* Componment - position - resolution - layer*/
		exit_res processCPRLsop();
		exit_res processCPRL();

		/* Use the information gathered to update the given marker segments */
		exit_res updateCOD(codClass &thisCod);
		exit_res updateSOT(sotClass &thisSot);

		// Markers to process
		// NOTE: SOC (start of codestream), SOD (start of data), EOC (end of codestream), and EPH (end of packet header)
		//		 contains only the marker, and needs no processing.

		// Delimiting markers and marker segments
		exit_res processSOT();

		// Fixed information marker segments
		exit_res processSIZ();

		// Functional marker segments
		exit_res processCOD();
		exit_res processCOC(); // optional
		exit_res processQCD(); 
		exit_res processQCC(); // optional
		exit_res processRGN();
		exit_res processPOC();

		// Pointer marker segments
		exit_res processPLM();
		exit_res processPLT();
		exit_res processPPM();
		exit_res processPPT();
		exit_res processTLM();

		// Informational marker segments
		exit_res processCRG();
		exit_res processCOM();

		// Handling not supported markers
		exit_res processUNKNOWN(uint16 marker);
};

} // End namespace js

#endif